#include <mist/mist.h>
#include <mist/io/image.h>
#include <mist/filter/linear.h>
#include <mist/filter/median.h>
#include <time.h>
#include <cmath>
#include <stdio.h>
#include "graph.h"

#define MASK 5
#define SHIFTMAX 255
#define SHIFTMIN 0
#define T 50
#define MEDIAN 3

int SAD( const int i, const int j, const int d, const mist::array2< int > &left, const mist::array2< int > &right )
{
  // SAD$B$N7W;;(B
  int sad;

  // $B=i4|2=(B
  sad = 0;
  if( i - d - MASK >= 0 && j > MASK && i + MASK < (int)left.width() && j + MASK < (int)left.height() )
    {
      // $B$H$j$"$($:CmL\2hAG<~JU$N9g7W$r7W;;(B
      for ( int m = - MASK; m <= MASK; m++ )
	for( int n = - MASK; n <= MASK; n++ )
	  {
	    
	    // $B:9$r7W;;$9$k(B SAD(sum. of absolue difference)$B$r;H$&(B		    
	    sad += abs( left( i + n , j + m ) - right( i - d + n, j + m ) );
	    //std::cout << " sad " << sad << std::endl; 
	    
	  }
      
      //  sad = abs( left( i, j) - right( i - d, j) );
    }

  return sad;

}

int main( int argc, char *argv[] )
{

  clock_t time1, time2;

  time1 = clock();

  // $B%=!<%Y%k%U%#%k%?$N=`Hw(B
  int fvd[ ] = 
    { 
      1,  1,  1, 
      0,  0,  0,
      -1, -1, -1 
    }; // $B=DJ}8~%(%C%8Cj=PMQ%^%9%/MWAG(B
  
  int fhd[ ] = 
    { 
      1, 0, -1, 
      1, 0, -1,
      1, 0, -1 
      }; // $B2#J}8~%(%C%8Cj=PMQ%^%9%/MWAG(B
  
  mist::array2< int > fv( 3, 3 );
  std::copy( fvd, fvd + 9, fv.begin( ) ); // fvd$B$NMWAG$r(Bfv$B$K%3%T!<(B
  
  mist::array2< int > fh( 3, 3 );
  std::copy( fhd, fhd + 9, fh.begin( ) ); // fhd$B$NMWAG$r(Bfh$B$K%3%T!<(B
  
  // $B2hA|$O%0%l!<%9%1!<%k7A<0$G07$&(B
  mist::array2< int > right, left;
  //mist::array2< int > t_right, t_left;

  int dispmax; // $B:GBg;k:9(B

  if( argc != 4 )
    {
      std::cout << "$B;H$$J}(B ./GraphCut $B:82hA|(B $B1&2hA|(B $B:GBg;k:9(B" << std::endl;
      return 1;
    }

  // bmp$B7A<0$N1&2hA|!$:82hA|$rFI$_9~$`(B
  // $B1&2hA|FI$_9~$_(B
  if( mist::read_bmp( right, argv[2] ) )
    {
      std::cout << " Right Image Reading Success " << std::endl;
    }
  else
    {
      std::cout << " Right Image Reading Failure " << std::endl;
      exit(0);
    }

  // $B:82hA|FI$_9~$_(B
  if( mist::read_bmp( left, argv[1] ) )
    {
      std::cout << " Left Image Reading Success " << std::endl;
    }
  else
    {
      std::cout << " Left Image Reading Failure " << std::endl;
      exit(0);
    }

  // $B:GBg;k:9FI$_9~$_(B
  dispmax = atoi(argv[3]);
  fprintf(stderr, "dispmax = %d\n", dispmax);

  // $BDI2C(B
//   mist::median( t_right, right, MEDIAN, 0 );
//   mist::median( t_left, left, MEDIAN, 0 );

  // $B%0%l!<%9%1!<%k$KJQ49$5$l$?1&2hA|$r=PNO(B 
  mist::write_bmp( right, "r_grayscale.bmp" );

  // $B%0%l!<%9%1!<%k$KJQ49$5$l$?:82hA|$r=PNO(B 
  mist::write_bmp( left, "l_grayscale.bmp" );

  // out:$B=PNO%U%!%$%k(B
  mist::array2< int > out( left.width(), left.height() ), tmp( left.width(), left.height() ); 

  // $B%a%G%#%"%s$H%=!<%Y%k%U%#%k%?MQ(B
  mist::array2< int > fx, fy, sobel, t_tmp( left.width(), left.height() );

  // $B%=!<%Y%k%U%#%k%?(B
  mist::linear_filter( left, fx, fh ); // fh$B$r%^%9%/$H$9$k@~7A%U%#%k%?!J2#J}8~(B1$B<!HyJ,!K(B
  mist::linear_filter( left, fy, fv ); // fv$B$r%^%9%/$H$9$k@~7A%U%#%k%?!J=DJ}8~(B1$B<!HyJ,!K(B
  
  sobel.resize( left.width( ), left.height( ) );
  
  for( size_t i = 0 ; i != sobel.size( ) ; ++ i )
    {
      sobel[ i ] = static_cast< int >( std::sqrt( ( fx[ i ] * fx[ i ] + fy[ i ] * fy[ i ] ) / 2.0 ) ); // $B=DJ}8~!$2#J}8~%(%C%86/EY9g@.(B
      
    }
  mist::write_bmp( sobel, "sobel.bmp" );
  

  int image_node_number, add_node_number, all_node_number, edge_number;

  // $B2hA|$rI=$9%N!<%I$N?t(B = $B2hA|$NI}(B * $B2hA|$N9b$5(B
  image_node_number = left.width() * left.height();

  // $B2hAG$H2hAG$N4V$KDI2C$9$k%N!<%I$N?t(B
  add_node_number = ( left.height() - 1 ) * left.width() + ( left.width() - 1 ) * left.height();
 
  // $B%N!<%I$N?t$N9g7W(B
  all_node_number = image_node_number + add_node_number;

  // $B%(%C%8$N?t(B
  edge_number = ( left.width() - 1 ) * 2 * left.height() + ( left.height() - 1 ) * 2 * left.width();

  fprintf(stderr, "node_number = %d\n", all_node_number);
  fprintf(stderr, "edge_number = %d\n", edge_number);

  
  /******************************$B&A3HD%$G%9%F%l%*(B****************************************/

  // $BJQ?t$N@k8@(B
  typedef Graph<int,int,int> GraphType;
  GraphType *g;

  // E = $B$H$F$bBg$-$JCM(B, k = $B%i%Y%k$N0LCV$r<($9(B
  int E = 1000000000, success, alpha, d_f, d_alpha, k, vfp_fq, va_a, vfp_a, va_fq, fp, fq, c;
  //int c0 = 2, c1 = 2;

  // $B=i4|CM(B
  for( size_t j = 0; j < left.height(); j++)
    for( size_t i = 0; i < left.width(); i++)
      {
	//tmp( i, j ) = 16;
	tmp( i, j ) = 0;
      }

  // $B$H$F$bBg$-$JCM$^$G%k!<%W(B
  for( int loop = 0; loop < E; loop++ )
    {
      success = 0;
   
      // for$B%k!<%W(B 0 $B$+$i(B dispmax $B$^$G%k!<%W(B
      for( alpha = 0; alpha <= dispmax; alpha++ )
	{	  

	  //fprintf(stderr, "loop %d - %d\n",loop, alpha);

	  // $B%0%i%U$N=i4|2=(B
	  g = new GraphType(/*estimated # of nodes*/ all_node_number, 
			    /*estimated # of edges*/ edge_number); 

	  // $B%N!<%I$N?t$V$s@_Dj(B
	  g -> add_node(all_node_number); 


	  /*******$B%9%F%l%*%^%C%A%s%0$N0Y$N%G!<%?%3%9%H4X?t$NDj5A(B*******/
	  // $BLdBj$O(Bk$B$NCM!%(B
	  // $B$H$j$"$($:<x6H;qNA$K1h$C$?7A(B

	  // $B=i4|2=(B 
	  k = 0; 

	  for( size_t j = 0; j < left.height(); j++)
	    for( size_t i = 0; i < left.width(); i++)
	      {
		// k$B$NA}$d$7J}$,LdBj!%(B
		// $B$?$V$s$3$l$G9g$C$F$k$H;W$&$s$@$1$I!&!&!&(B
		k = j * ( left.width() + left.width() - 1 + left.width() ) + i * 2;

		// SAD$B$N7W;;(B
		d_f = SAD( i, j, tmp( i, j ), left, right );
		d_alpha = SAD( i, j, alpha, left, right );

		//fprintf(stderr, "%d \n", d_ap);

		d_f = abs( d_f );
		d_alpha = abs( d_alpha );

		//fprintf(stderr, "%d \n", d_ap);
						
		// $B%N!<%I$N!$(BSOURCE$BB&$K(BD(fp),SINK$BB&$K(BD($B&A(B)$B$r@_Dj(B
		// $B:G=*E*$K(Bsad$B$N7W;;7k2L$,%i%Y%k$K$J$k$N$+(B
		g -> add_tweights( k,   /* capacities */  d_f, d_alpha );
	
	      }
	  /****************************************************/	  



	  /********$B%9%F%l%*%^%C%A%s%0$N0Y$N%9%`!<%:%3%9%H$NDj5A(B********/

	  // $B=i4|2=(B
	  k = 0; 

	  for( size_t j = 0; j < left.height(); j++)
	    for( size_t i = 0; i < left.width() - 1; i++)
	      {
		// k$B$O$3$l$G$$$$$O$:!&!&!&(B
		k = j * ( left.width() + left.width() - 1 + left.width() ) + i * 2 + 1;

		// $B@:EY8~>e$N0Y$K(B
		if( sobel( i, j ) > T )
		  c = 1;
		else
		  c = 5;		

		//c = 1;

		// A[alpha][]$B$N$&$A!$&B(Bp$B$K:G$b6a$$CM$r&A(Bp$B$K@_Dj(B
		// A[alpha][]$B$N$&$A!$&B(Bq$B$K:G$b6a$$CM$r&A(Bq$B$K@_Dj(B
		fp = tmp( i, j );
		fq = tmp( i+1, j );

		// SINK$BB&$O6&DL(B
		va_a = c * abs( alpha - alpha );

		// $B2#(B
		vfp_fq = c * abs( fp - fq );
		
		// $B%N!<%I(Ba$B$N!$(BSOURCE$BB&$K(BV(bp, bq), SINK$BB&$K(BV($B&A(B, $B&A(B)$B$r@_Dj(B
		g -> add_tweights( k,   /* capacities */  vfp_fq, va_a );

		// $B:8NY(B
		vfp_a = abs( alpha - fp );

		// $B%N!<%I(Bp$B$H%N!<%I(Ba$B$N%(%C%8$N=E$_(B($BAPJ}8~(B)$B$K(BV(fp,$B&A(B)$B$r@_Dj(B
		g -> add_edge( k - 1, k,   /* capacities */ vfp_a, vfp_a );

		// $B1&NY(B
		va_fq = abs( alpha - fq );

		// $B%N!<%I(Bq$B$H%N!<%I(Ba$B$N%(%C%8$N=E$_(B($BAPJ}8~(B)$B$K(BV($B&A(B, fq)$B$r@_Dj(B
		g -> add_edge( k, k + 1,   /* capacities */ va_fq, va_fq );
			

	      }
	  
	  
	  k = 0;
	  for( size_t j = 0; j < left.height() - 1; j++)
	    for( size_t i = 0; i < left.width(); i++)
	      {
		// k$B$O$3$l$G$$$$$O$:!&!&!&(B
		k = ( j + 1 ) * ( left.width() + left.width() - 1 ) + i + ( left.width() * j );

		// $B@:EY8~>e$N0Y$K(B
		if( sobel( i, j ) > T )
		  c = 1;
		else
		  c = 5; 

		//c = 1;
		
		/*
		  fprintf(stderr, "k = %d\n", k);
		  fprintf(stderr, "$B>e(B = %d\n",j * ( in.width() + in.width() - 1 + in.width() ) + i * 2);
		  fprintf(stderr, "$B2<(B = %d\n",( j + 1 ) * ( in.width() + in.width() - 1 + in.width() ) + i * 2);
		*/

		fp = tmp( i, j );
		fq = tmp( i, j + 1 );

		// SINK$BB&$O6&DL(B
		va_a = c * abs( alpha - alpha );

		// $B=D(B
		vfp_fq = c * abs( fp - fq );
		
		// $B%N!<%I(Ba$B$N!$(BSOURCE$BB&$K(BV(fp, fq), SINK$BB&$K(BV($B&A(B, $B&A(B)$B$r@_Dj(B
		g -> add_tweights( k,   /* capacities */  vfp_fq, va_a );

		// $BCmL\%(%C%8$H>e(B
		vfp_a = abs( alpha - fp );

		// $B%N!<%I(Bp$B$H%N!<%I(Ba$B$N%(%C%8$N=E$_(B($BAPJ}8~(B)$B$K(BV(fp,$B&A(B)$B$r@_Dj(B
		g -> add_edge( j * ( left.width() + left.width() - 1 + left.width() ) + i * 2, k,   /* capacities */ vfp_a, vfp_a );

		// $BCmL\%(%C%8$H2<(B
		va_fq = abs( alpha - fq );

		// $B%N!<%I(Bq$B$H%N!<%I(Ba$B$N%(%C%8$N=E$_(B($BAPJ}8~(B)$B$K(BV($B&A(B, fq)$B$r@_Dj(B
		g -> add_edge( k, ( j + 1 ) * ( left.width() + left.width() - 1 + left.width() ) + i * 2,   /* capacities */ va_fq, va_fq );    	

	      }
	
      
	  /**************************************************************************/	   
      

	  //fprintf(stderr,"graph_definition\n");
	  // $B:GBgN.!&:G>.%+%C%H%"%k%4%j%:%`$NE,MQ(B
	  int flow = g -> maxflow();
      
	  // flow < E $B$J$i8=:_$N%i%Y%k$r5a$^$C$?%i%Y%k$K$7!$(BE = flow$B$K$7!$(Bsuccess = 1$B$K$9$k(B
	  if( flow < E )
	    {
	      fprintf(stderr, "alpha = %d\n", alpha);
	      // flow : $B5a$^$C$?%i%Y%k$G7W;;$7$?Am%3%9%H4X?t(B($B$?$V$s(B)  
	      fprintf(stderr,"Flow = %d\n", flow);
	      
	      
	      E = flow;
	      
	      for( size_t j = 0; j < left.height(); j++)
		for( size_t i = 0; i < left.width(); i++)
		  {

		    k = j * ( left.width() + left.width() - 1 + left.width() ) + i * 2;
		    
		    if (g->what_segment(k) == GraphType::SOURCE)
		      {
			// printf("tmp(%d, %d) = node%d is in the SOURCE set\n", i, j, k);
			tmp( i, j ) = alpha;
			//fprintf(stderr, "tmp = %d\n", tmp(i,j));
		      }
		  }

	      // $B%a%G%#%"%s%U%#%k%?$G%N%$%:=|5n(B
	      mist::median( tmp, t_tmp, MEDIAN, 0 );

	      for( size_t j = 0; j < left.height(); j++ )
		for( size_t i = 0; i < left.width(); i++ )
		  {
		    tmp( i, j ) = t_tmp( i, j );
		  }	   
	      success = 1;
	    }

	  // $B%0%i%U$N>C5n(B
	  delete g;
	}

      // success == 0$B$J$i(Bbreak;
      if( success == 0 ) 
	break;
  
    }
  
  // bmp$B7A<0$N2hA|%U%!%$%k$r=PNO(B 
  mist::write_bmp( tmp, "tmp.bmp" );

  mist::array2< double > shift( left.width(), left.height() );
  int max = 0, min = 1000;

  // $BG;EY3,D4JQ49(B

  for( size_t j = 0; j < tmp.height(); j++ )
    {
      for( size_t i = 0; i < tmp.width(); i++ )
	{
	  if( max < tmp( i, j ) )
	    {
	      max = tmp( i, j );
	    }
	  if( min > tmp( i, j ) )
	    {
	      min = tmp( i, j );
	    }
	}
    }


  for( size_t j = 0; j < tmp.height(); j++ )
    {
      for( size_t i = 0; i < tmp.width(); i++ )
	{
	  if( tmp( i, j ) < min )
	    shift( i, j ) = SHIFTMIN;
	  else if( min <= tmp( i, j ) && tmp( i, j ) <= max )
	    {
	      shift( i, j ) = ( SHIFTMAX / ( (double)( max - min ) ) * ( (double)( tmp( i, j ) - min ) ) );
	    }
	  else if( max < tmp( i, j ) )
	    shift( i, j ) = SHIFTMAX; 
	}
    }

  for( size_t j = 0; j < out.height(); j++ )
    for( size_t i = 0; i < out.width(); i++ )
      {
	out( i, j ) = ( int )shift( i, j );
      }
  
  mist::write_bmp( out, "out.bmp" );
  
  time2 = clock();

  std::cout << "$B7W;;;~4V(B: " << (double)(time2 - time1) / CLOCKS_PER_SEC << std::endl;

  return 0;
}
